/******************************************************************************/
/*                                                                            */
/* I N T E R N A T I O N A L  D A T A  E N C R Y P T I O N  A L G O R I T H M */
/*                                                                            */
/*                     A S   A   U S E R   C O M M A N D                      */
/*                                                                            */
/******************************************************************************/
/* Author:       Richard De Moliner (demoliner@isi.ee.ethz.ch)                */
/*               Signal and Information Processing Laboratory                 */
/*               Swiss Federal Institute of Technology                        */
/*               CH-8092 Zuerich, Switzerland                                 */
/* Created:      April 23, 1992                                               */
/* Changes:      November 16, 1993 (support of ANSI-C and C++)                */
/* System:       SUN SPARCstation, SUN acc ANSI-C-Compiler, SUN-OS 4.1.3      */
/******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "idea.h"

#define TRUE                 1 /* boolean constant for true                   */
#define FALSE                0 /* boolean constant for false                  */
#define nofTestData     163840 /* number of blocks encrypted in time test     */

#define nomode               0 /* no mode is specified                        */
#define ecb                  1 /* electronic code book mode                   */
#define cbc                  2 /* cipher block chaining mode                  */
#define cfb                  3 /* ciphertext feedback mode                    */
#define ofb                  4 /* output feedback mode                        */
#define eol              0x100 /* end of line character                       */
#define colon            0x101 /* character ':'                               */
#define error            0x102 /* unknown character                           */
#define maxInterleave     1024 /* maximal interleave factor + 1               */
#define nofChar ('~' - '!' +1) /* number of different printable characters    */
#define maxBufLen (Idea_dataSize * 1024) /* size of input and output buffer   */

Idea_UserKey userKey;          /* user selected 128 bit key                   */
Idea_Key key;                  /* expanded key with 832 bits                  */
Idea_Data state[maxInterleave];/* state informations for interleaving modes   */

u_int32 inputLen    = 0;       /* current number of bytes read from 'inFile'  */
int interleave      = 0;       /* current interleave factor                   */
int time_0          = 0;       /* time for interleaving modes                 */
int time_N          = 0;       /* time-interleave for interleaving modes      */
int mode            = nomode;  /* current mode                                */

int optEncrypt      = FALSE;   /* encrypt option 'e'                          */
int optDecrypt      = FALSE;   /* decrypt option 'd'                          */
int optKeyHexString = FALSE;   /* key as hex-string option 'K'                */
int optKeyString    = FALSE;   /* key as string option 'k'                    */

int inBufLen        = maxBufLen; /* current length of data in 'inBuf'         */
int inBufPos        = maxBufLen; /* current read position of 'inBuf'          */
int outBufLen       = 0;       /* current write position of 'outBuf'          */
u_int8 inBuf[maxBufLen];       /* buffer for file read                        */
u_int8 outBuf[maxBufLen];      /* buffer for file write                       */

FILE *inFile;                  /* file with input data (plain or ciphertext)  */
FILE *outFile;                 /* file for output data (plain or ciphertext)  */

/******************************************************************************/
/* initialize global variables                                                */

void Init(void)
{ int i, pos;

  for (i = Idea_userKeyLen - 1; i >= 0 ; i--) userKey[i] = 0;
  for (pos = maxInterleave - 1; pos >= 0 ; pos--)
    for (i = Idea_dataLen - 1; i >= 0; i--)
      state[pos][i] = 0;
} /* Init */

/******************************************************************************/
/*                          E R R O R - H A N D L I N G                       */
/******************************************************************************/
/* write usage error message and terminate program                            */

void UsageError(int num)
{
fprintf(stderr, "(%d)\n\
Usage:   idea [ -e | -d ] [ -ecb | -cbcN | -cfbN | -ofbN ]       \n\
              ( -k keyString | -K keyHexString )                 \n\
              [ inputFile [ outputFile ] ]                       \n\
\nExample: idea -e -k \"k e y\" infile | idea -cbc8 -K 123:9a::eF - outfile\n\
\n", num);
  exit(-1);
} /* UsageError */

/******************************************************************************/
/* write error message and terminate program                                  */

void Error(int num, char *str)
{  fprintf(stderr, "error %d in idea: %s\n", num, str); exit(-1); } /* Error */

/******************************************************************************/
/* write system error message and terminate program                           */

void PError(char *str)
{ perror(str); exit(-1); } /* PError */

/******************************************************************************/
/*                          D E C R Y P T I O N  /  E N C R Y P T I O N       */
/******************************************************************************/
/* read one data-block from 'inFile'                                          */

int GetData(Idea_Data data)
{ register int i, len;
  register u_int16 h;
  register u_int8 *inPtr;

  if (inBufPos >= inBufLen) {
    if (inBufLen != maxBufLen) return 0;
    inBufLen = fread(inBuf, 1, maxBufLen, inFile);
    inBufPos = 0;
    if (inBufLen == 0) return 0;
    if (inBufLen % Idea_dataSize != 0)
      for (i = inBufLen; i % Idea_dataSize != 0; i++) inBuf[i] = 0;
  }
  inPtr = &inBuf[inBufPos];
  for (i = 0; i < Idea_dataLen; i++) {
    h = ((u_int16)*inPtr++ & 0xFF) << 8;
    data[i] = h | (u_int16)*inPtr++ & 0xFF;
  }
  inBufPos += Idea_dataSize;
  if (inBufPos <= inBufLen) len = Idea_dataSize;
  else len = inBufLen + Idea_dataSize - inBufPos;
  inputLen += len;
  return len;
} /* GetData */

/******************************************************************************/
/* write one data-block to 'outFile'                                          */

void PutData(Idea_Data data, int len)
{ register int i;
  register u_int16 h;
  register u_int8 *outPtr;

  outPtr = &outBuf[outBufLen];
  for (i = 0; i < Idea_dataLen; i++) {
    h = data[i];
    *outPtr++ = h >> 8 & 0xFF;
    *outPtr++ = h & 0xFF;
  }
  outBufLen += len;
  if (outBufLen >= maxBufLen) {
    fwrite(outBuf, 1, maxBufLen, outFile);
    outBufLen = 0;
  }
} /* PutData */

/******************************************************************************/
/* write last block to 'outFile' and close 'outFile'                          */

void CloseOutput(void)
{ if (outBufLen > 0) {
    fwrite(outBuf, 1, outBufLen, outFile);
    outBufLen = 0;
  }
  fclose(outFile);
} /* CloseOutput */

/******************************************************************************/
/* increment time_0 and time_N                                                */

void IncTime(void)
{ time_0 = (time_0 + 1) % maxInterleave;
  time_N = (time_N + 1) % maxInterleave;
} /* IncTime */

/******************************************************************************/
/* encrypt one data-block                                                     */

void EncryptData(Idea_Data data)
{ int i;

  switch (mode) {
    case ecb:
      Idea_Crypt(data, data, key);
      break;
    case cbc:
      for (i = Idea_dataLen - 1; i >= 0; i--) data[i] ^= state[time_N][i];
      Idea_Crypt(data, data, key);
      for (i = Idea_dataLen - 1; i >= 0; i--) state[time_0][i] = data[i];
      IncTime();
      break;
    case cfb:
      Idea_Crypt(state[time_N], state[time_0], key);
      for (i = Idea_dataLen - 1; i >= 0; i--)
        data[i] = state[time_0][i] ^= data[i];
      IncTime();
      break;
    case ofb:
      Idea_Crypt(state[time_N], state[time_0], key);
      for (i = Idea_dataLen - 1; i >= 0; i--) data[i] ^= state[time_0][i];
      IncTime();
      break;
    default: break;
  }
} /* EncryptData */

/******************************************************************************/
/* decrypt one data-block                                                     */

void DecryptData(Idea_Data data)
{ int i;

  switch (mode) {
    case ecb:
      Idea_Crypt(data, data, key);
      break;
    case cbc:
      for (i = Idea_dataLen - 1; i >= 0; i--) state[time_0][i] = data[i];
      Idea_Crypt(data, data, key);
      for (i = Idea_dataLen - 1; i >= 0; i--) data[i] ^= state[time_N][i];
      IncTime();
      break;
    case cfb:
      for (i = Idea_dataLen - 1; i >= 0; i--) state[time_0][i] = data[i];
      Idea_Crypt(state[time_N], data, key);
      for (i = Idea_dataLen - 1; i >= 0; i--) data[i] ^= state[time_0][i];
      IncTime();
      break;
    case ofb:
      Idea_Crypt(state[time_N], state[time_0], key);
      for (i = Idea_dataLen - 1; i >= 0; i--) data[i] ^= state[time_0][i];
      IncTime();
      break;
    default: break;
  }
} /* DecryptData */

/******************************************************************************/
/* store integer 'value' in 'data'                                            */

void PlainLenToData(u_int32 value, Idea_Data data)
{ data[3] = (u_int16)(value << 3 & 0xFFFF);
  data[2] = (u_int16)(value >> 13 & 0xFFFF);
  data[1] = (u_int16)(value >> 29 & 0x0007);
  data[0] = 0;
} /* PlainLenToData */

/******************************************************************************/
/* extract integer 'value' from 'data'                                        */

void DataToPlainLen(Idea_Data data, u_int32 *value)
{ if (data[0] || data[1] > 7 || data[3] & 7)
    Error(0, "input is not a valid cryptogram");
  *value = (u_int32)data[3] >> 3 & 0x1FFF |
           (u_int32)data[2] << 13 |
           (u_int32)data[1] << 29;
} /* DataToPlainLen */

/******************************************************************************/
/* encrypt / decrypt complete data-stream or compute hash value of data-stream*/

void CryptData(void)
{ int t, i;
  u_int32 len;
  Idea_Data dat[4];
  Idea_Data data;

  if (optEncrypt) { /* encrypt data */
    while ((len = GetData(data)) == Idea_dataSize) {
      EncryptData(data); 
      PutData(data, Idea_dataSize); 
    }
    if (len) { EncryptData(data); PutData(data, Idea_dataSize); }
    PlainLenToData(inputLen, data);
    EncryptData(data);
    PutData(data, Idea_dataSize);
    CloseOutput();
  }
  else if (optDecrypt) { /* decrypt data */
    if ((len = GetData(dat[0])) != Idea_dataSize) {
      if (len) Error(2, "input is not a valid cryptogram");
      else Error(3, "there are no data to decrypt");
    }
    DecryptData(dat[0]);
    if ((len = GetData(dat[1])) != Idea_dataSize) {
      if (len) Error(4, "input is not a valid cryptogram");
      DataToPlainLen(dat[0], &len);
      if (len) Error(5, "input is not a valid cryptogram");
    }
    else {
      DecryptData(dat[1]);
      t = 2;
      while ((len = GetData(dat[t])) == Idea_dataSize) {
        DecryptData(dat[t]);
        PutData(dat[(t + 2) & 3], Idea_dataSize);
        t = (t + 1) & 3;
      }
      if (len) Error(6, "input is not a valid cryptogram");
      DataToPlainLen(dat[(t + 3) & 3], &len);
      len += 2 * Idea_dataSize;
      if (inputLen < len && len <= inputLen + Idea_dataSize) {
        len -= inputLen;
        PutData(dat[(t + 2) & 3], len);
      }
      else Error(7, "input is not a valid cryptogram");
    }
    CloseOutput();
  }
} /* CryptData */

/******************************************************************************/
/*                          I N I T I A L I Z A T I O N                       */
/******************************************************************************/
/* set option to TRUE                                                         */

void SetOption(int *option)
{ if (*option) UsageError(10);
  *option = TRUE;
} /* SetOption */

/******************************************************************************/
/* set encryption / decryption mode                                           */

void SetMode(int newMode, char **str)
{ if (mode != nomode) UsageError(11);
  mode = newMode;
  (*str)++; (*str)++;
  if (newMode == cbc || newMode == cfb || newMode == ofb) {
    if ('0' <= **str && **str <= '9') {
      interleave = 0;
      do {
        interleave = 10 * interleave + (**str - '0');
        if (interleave >= maxInterleave)
          Error(12, "interleave factor is too large");
        (*str)++;
      } while ('0' <= **str && **str <= '9');
      if (interleave == 0) Error(13, "interleave factor is zero");
    }
    else interleave = 1;
  }
} /* SetMode */

/******************************************************************************/
/* read options from string 'str'                                             */

void ReadOptions(char *str, int *readKeyString, int *readKeyHexString)
{ char ch;

  str++;
  *readKeyString = *readKeyHexString = FALSE;
  while((ch = *str++) != '\0') {
    switch (ch) {
      case 'a':
        UsageError(14);
        break;
      case 'c':
        if (str[0] == 'b' && str[1] == 'c') SetMode(cbc, &str);
        else if (str[0] == 'f' && str[1] == 'b') SetMode(cfb, &str);
        else UsageError(15);
        break;
      case 'd': SetOption(&optDecrypt); break;
      case 'e': 
        if (str[0] == 'c' && str[1] == 'b') SetMode(ecb, &str);
        else SetOption(&optEncrypt);
        break;
      case 'o':
        if (str[0] == 'f' && str[1] == 'b') SetMode(ofb, &str);
        else UsageError(16);
        break;
      case 'k': SetOption(&optKeyString); *readKeyString = TRUE; break;
      case 'K': SetOption(&optKeyHexString); *readKeyHexString = TRUE; break;
      default: UsageError(18); break;
    }
  }
} /* ReadOptions */

/******************************************************************************/
/* check if options are unique and set default options                        */

void AdjustOptions(void)
{
  {
    if (optDecrypt && optEncrypt) UsageError(21);
    if (optKeyString && optKeyHexString) UsageError(23);
    if (!optDecrypt && !optEncrypt)
      SetOption(&optEncrypt);
    else {
      if (mode == nomode) { mode = cbc; interleave = 1; }
      else if (mode != ecb && mode != cbc && mode != cfb && mode != ofb)
        UsageError(27);
      if (!optKeyString && !optKeyHexString) UsageError(28);
    }
    time_0 = interleave;
    time_N = 0;
  }
} /* AdjustOptions */

/******************************************************************************/
/* convert a hex-digit into an integer                                        */

u_int32 HexToInt(char ch)
{ if ('0' <= ch && ch <= '9') return ch - '0';
  else if ('a' <= ch  && ch <= 'f') return 10 + (ch - 'a');
  else if ('A' <= ch && ch <= 'F') return 10 + (ch - 'A');
  else if (ch == ':') return colon;
  else if (ch == '\0') return eol;
  else return error;
} /* HexToInt */

/******************************************************************************/
/* convert a character into an integer                                        */

u_int32 CharToInt(char ch)
{ if ('!' <= ch && ch <= '~') return ch - '!';
  else if (ch == '\0') return eol;
  else return error;
} /* CharToInt */

/******************************************************************************/
/* initializes key and initial values                                         */

void ReadKeyHexString(char *str)
{ int pos, i;
  u_int32 val;

  while ((val = HexToInt(*str++)) < eol) {
    for (i = Idea_userKeyLen - 1; i >= 0; i--) {
      val |= (u_int32)userKey[i] << 4;
      userKey[i] = (u_int16)(val & 0xFFFF);
      val >>= 16;
    }
    if (val) Error(29, "key value is too large");
  }
  for (pos = 0; val == colon && pos < maxInterleave; pos++) {
    while ((val = HexToInt(*str++)) < eol) {
      for (i = Idea_dataLen - 1; i >= 0; i--) {
        val |= (u_int32)state[pos][i] << 4;
        state[pos][i] = (u_int16)(val & 0xFFFF);
        val >>= 16;
      }
      if (val) Error(30, "initial value is too large");
    }
  }
  if (val == colon) Error(31, "too many initial values specified");
  if (val != eol) Error(32, "wrong character in initialization string");
} /* ReadKeyHexString */

/******************************************************************************/
/* initialize key and initial values                                          */

void ReadKeyString(char *str)
{ int i;
  u_int32 val;

  while ((val = CharToInt(*str++)) < eol) {
    for (i = Idea_userKeyLen - 1; i >= 0; i--) {
      val += (u_int32)userKey[i] * nofChar;
      userKey[i] = (u_int16)(val & 0xFFFF);
      val >>= 16;
    }
  }
  if (val != eol) Error(32, "wrong character in key string");
} /* ReadKeyString */

/******************************************************************************/
/* show current state informations                                            */

void ShowState(void)
{ int i, j;

  fprintf(stderr, "Mode = {");
  switch (mode) {
    case ecb: fprintf(stderr, "ecb"); break;
    case cbc: fprintf(stderr, "cbc"); break;
    case cfb: fprintf(stderr, "cfb"); break;
    case ofb: fprintf(stderr, "ofb"); break;
    case nomode: fprintf(stderr, "nomode"); break;
    default: fprintf(stderr, "!!!wrong mode!!!"); break;
  }
  if (interleave > 0) fprintf(stderr, "%d", interleave);
  if (optEncrypt) fprintf(stderr, ", encrypt");
  if (optDecrypt) fprintf(stderr, ", decrypt");
  if (optKeyString) fprintf(stderr, ", key string");
  if (optKeyHexString) fprintf(stderr, ", key hex string");
  fprintf(stderr, "}\n\nKey:\n");
  for (i = 0; i < Idea_keyLen; i++) {
    fprintf(stderr, "%7u<%4x>", key[i], key[i]);
    if ((i % 6) == 5) fprintf(stderr, "\n");
  }
  fprintf(stderr, "\n\nInitial values:");
  for (i = 0; i < interleave; i++) {
    fprintf(stderr, "\n  x[N -%2d] =", i + 1);
    for (j = 0; j < Idea_dataLen; j++)
      fprintf(stderr, "%7u<%4x>", state[i][j], state[i][j]);
  }
  fprintf(stderr, "\n");
} /* ShowState */

/******************************************************************************/
/*                          M A I N - P R O C E D U R E                       */
/******************************************************************************/

int main(int argc, char *argv[])
{ int readKeyString, readKeyHexString;

 Init();
 argv++; argc--;
 while (argc > 0 && argv[0][0] == '-' && argv[0][1] != '\0') {
   ReadOptions(*argv++, &readKeyString, &readKeyHexString); argc--;
   if (readKeyString || readKeyHexString) {
     if (argc <= 0)  Error(36, "missing key on command line");
     else if (readKeyString) { ReadKeyString(*argv++); argc--; }
     else { ReadKeyHexString(*argv++); argc--; }
   }
 }
 AdjustOptions();
  {
    if (argc > 1 && strcmp(argv[0], argv[1]) == 0)
      Error(38, "source and destination are identical");
    if (argc > 2 && strcmp(argv[0], argv[2]) == 0)
      Error(39, "source and destination are identical");
    if (argc > 2 && strcmp(argv[1], argv[2]) == 0)
      Error(40, "destinations are identical");
    inFile = stdin;
    outFile = stdout;
    if (argc > 0) {
      if (strcmp(*argv, "-") == 0) { argv++; argc--; }
      else {
        inFile = fopen(*argv++, "rb"); argc--;
        if (inFile == NULL) PError(*--argv);
      }
    }
    {
      if (argc > 0) {
        outFile = fopen(*argv++, "wb"); argc--;
        if (outFile == NULL) PError(*--argv);
      }
    }
    if (argc > 0) Error(41, "too many parameters");
    Idea_ExpandUserKey(userKey, key);
    if (optDecrypt && (mode == ecb || mode == cbc)) Idea_InvertKey(key, key);
    CryptData();
  }
  return 0;
}
